"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSDataTemplate = exports.TSHTMLTemplate = exports.TSGenericDataTemplate = exports.TSStringTemplate = exports.TSFusionTemplate = exports.TSCharRole = void 0;
const commons_1 = require("./commons");
const compare_1 = require("./compare");
const data_1 = require("./data");
const tscharset_1 = require("./tscharset");
const tsdata_1 = require("./tsdata");
const tserrors_1 = require("./tserrors");
const tsfusionnode_1 = require("./tsfusionnode");
const utils_1 = require("./utils");
var TSCharRole;
(function (TSCharRole) {
    TSCharRole[TSCharRole["UnsignificantStarterChar"] = 0] = "UnsignificantStarterChar";
    TSCharRole[TSCharRole["StarterChar"] = 1] = "StarterChar";
    TSCharRole[TSCharRole["InternalChar"] = 2] = "InternalChar";
    TSCharRole[TSCharRole["ForbiddenChar"] = 3] = "ForbiddenChar";
})(TSCharRole = exports.TSCharRole || (exports.TSCharRole = {}));
;
const [LF, CR, TAB, SPACE, EXCLAMATION_MARK, DOUBLE_QUOTE, SHARP, DOLLAR, QUOTE, ASTERISK, PLUS, MINUS, DOT, SLASH, ZERO, NINE, COLON, LT, EQ, GT, QUESTION_MARK, AT, A, Z, UNDERSCORE] = [0x0a, 0x0d, 0x09, 0x20, 0x21, 0x22, 0x23, 0x24, 0x27, 0x2A, 0x2B, 0x2D, 0x2E, 0x2F, 0x30, 0x39, 0x3A, 0x3C, 0x3D, 0x3E, 0x3F, 0x40, 0x41, 0x5A, 0x5F];
const FUSION_SPECIAL_CHARS = [SHARP, QUESTION_MARK, EXCLAMATION_MARK, DOLLAR, AT, UNDERSCORE, ASTERISK, PLUS, DOT];
const TSFusionNodeTypes = [];
TSFusionNodeTypes[SHARP] = tsfusionnode_1.TSFusionNodeType.Enumeration;
TSFusionNodeTypes[QUESTION_MARK] = tsfusionnode_1.TSFusionNodeType.Test;
TSFusionNodeTypes[EXCLAMATION_MARK] = tsfusionnode_1.TSFusionNodeType.NegativeTest;
const TSFusionContextTypes = [];
TSFusionContextTypes[DOLLAR] = tsfusionnode_1.TSFusionContextType.Global;
TSFusionContextTypes[PLUS] = tsfusionnode_1.TSFusionContextType.User;
TSFusionContextTypes[AT] = tsfusionnode_1.TSFusionContextType.Root;
TSFusionContextTypes[UNDERSCORE] = tsfusionnode_1.TSFusionContextType.System;
TSFusionContextTypes[ASTERISK] = tsfusionnode_1.TSFusionContextType.Procedure;
TSFusionContextTypes[DOT] = tsfusionnode_1.TSFusionContextType.Local;
const TSFusionForbiddenEncodings = ['utf16le', 'base64', 'base64url', 'hex'];
class TSFusionTemplate {
    static fromData(source, encoding, opts) { return TSDataTemplate.fromData(source, encoding, opts); }
    static fromHTMLData(source, opts) { return TSHTMLTemplate.fromHTMLData(source, opts); }
    static fromString(source, opts) { return TSStringTemplate.fromString(source, opts); }
    constructor(src, opts = {}) {
        this._capacity = 0;
        this._procedures = {};
        this._variables = new Map();
        this._contextsVariables = new Map();
        const sm = _ascii2data((0, commons_1.$length)(opts.startingMark) ? opts.startingMark : TSFusionTemplate.DefaultStartingMark);
        const em = _ascii2data((0, commons_1.$length)(opts.endingMark) ? opts.endingMark : TSFusionTemplate.DefaultEndingMark);
        const sepa = _ascii2data((0, commons_1.$length)(opts.separator) ? opts.separator : TSFusionTemplate.DefaultSeparator);
        const debug = !!opts.debugParsing;
        const source = src instanceof ArrayBuffer ? (0, data_1.$bufferFromArrayBuffer)(src) : (src instanceof tsdata_1.TSData ? src.mutableBuffer : src);
        if (debug) {
            (0, utils_1.$logterm)(`&0\n&Y&k ${this.constructor.name} parsing in debug mode &0`);
            (0, utils_1.$logterm)(`&0&x - &w starting mark&x:        '&o${_data2ascii(sm)}&x'&0`);
            (0, utils_1.$logterm)(`&0&x - &w ending mark&x:          '&o${_data2ascii(em)}&x'&0`);
            (0, utils_1.$logterm)(`&0&x - &w separator&x:            '&o${_data2ascii(sepa)}&x'&0`);
            (0, utils_1.$logterm)('&0&x======== &wsource&x: ========&0\n&y' + (0, utils_1.$inspect)(source) + "&0");
        }
        if ((0, compare_1.$equal)(sm, em) || (0, compare_1.$equal)(sepa, sm) || (0, compare_1.$equal)(sepa, em)) {
            throw new tserrors_1.TSError('starting mark, ending mark and seperator must be distinct', { startingMark: sm, endingMark: em, separator: sepa });
        }
        const len = source.length;
        const smlen = sm.length;
        const sepalen = sepa.length;
        const emlen = em.length;
        function upper(c) { return (c & 0xdf); }
        function isVariableSpace(c) { return c === TAB || c === SPACE; }
        function variableCharacterRole(c) {
            if (c === DOT) {
                return TSCharRole.UnsignificantStarterChar;
            }
            if ((upper(c) >= A && upper(c) <= Z)) {
                return TSCharRole.StarterChar;
            }
            if ((c >= ZERO && c <= NINE) || c === DOT || c === UNDERSCORE) {
                return TSCharRole.InternalChar;
            }
            return TSCharRole.ForbiddenChar;
        }
        function store_variable(m, cm, variable, context) {
            const path = variable.split('.');
            for (let component of path) {
                if (component.length > 0) {
                    let cset = m.get(component);
                    if (!(0, commons_1.$ok)(cset)) {
                        cset = new Set();
                        m.set(component, cset);
                    }
                    cset.add(context);
                    let sset = cm.get(context);
                    if (!(0, commons_1.$ok)(sset)) {
                        sset = new Set();
                        cm.set(context, sset);
                    }
                    sset.add(component);
                    break;
                }
            }
        }
        function tohex(v) { return (0, commons_1.$isnumber)(v) ? `0x${v.toString(16)}` : `'${v}'`; }
        function tochar(c) { return String.fromCharCode(c); }
        if (debug) {
            (0, utils_1.$logterm)(`&0&x - &w source length&x:         &o${len}&0`);
            (0, utils_1.$logterm)(`&0&x - &w starting mark length&x:  &o${smlen}&0`);
            (0, utils_1.$logterm)(`&0&x - &w ending mark length&x:    &o${emlen}&0`);
            (0, utils_1.$logterm)(`&0&x - &w separator length&x:      &o${sepalen}&0`);
        }
        if (!len || !smlen || !sepalen || !emlen || smlen + emlen + sepalen > len) {
            throw new tserrors_1.TSError('template string to small', { startingMark: sm, endingMark: em, separator: sepa, source: source });
        }
        FUSION_SPECIAL_CHARS.forEach(c => {
            if (sm.includes(c) || em.includes(c) || sepa.includes(c)) {
                throw new tserrors_1.TSError(`starting mark, ending mark and seperator must not containe ${tohex(c)} ('${tochar(c)}') character`, { startingMark: sm, endingMark: em, separator: sepa, character: c });
            }
        });
        if (sm[0] === em[0] || sm[0] === sepa[0] || em[0] === sepa[0]) {
            throw new tserrors_1.TSError('first character of starting mark, ending mark and seperator must be distinct', { startingMark: sm, endingMark: em, separator: sepa });
        }
        let State;
        (function (State) {
            State["Text"] = "Text";
            State["OpenVariable"] = "open-var";
            State["PreVariableSpace"] = "pre-var-space";
            State["StartDecodingVariable"] = "start-decoding-var";
            State["DecodingVariable"] = "decoding-var";
            State["DecodeEndingMark"] = "decoding-end-mark";
            State["PostVariableSpace"] = "post-var-space";
            State["DecodingClosingVariable"] = "decoding-closing-var";
            State["DecodingSeparator"] = "decoding-separator";
        })(State || (State = {}));
        ;
        let state = State.Text;
        let tokenPos = 0;
        let characterRole;
        const root = tsfusionnode_1.TSFusionTreeNode.root();
        let current = root;
        let last = 0;
        let beginText = 0;
        let fusionContext = tsfusionnode_1.TSFusionContextType.Local;
        let i = 0;
        let comp;
        while (i < len) {
            const c = source[i];
            if (debug) {
                (0, utils_1.$logterm)(`&0&g${i.fpad(8, '_')}&w: &x&o${tohex(c)}&x (&b'&c${tochar(c)}&b'&x), state:&o${state}&x, current: &p${current.label}&x/&y${current.type}&0`);
            }
            switch (state) {
                case State.Text:
                    if (c === sm[0] || c === em[0]) {
                        last = i;
                        tokenPos = 0;
                        state = c === sm[0] ? State.OpenVariable : State.DecodeEndingMark;
                        i--;
                    }
                    break;
                case State.OpenVariable:
                    comp = sm[tokenPos++];
                    if (c !== comp) {
                        tokenPos = 0;
                        state = State.Text;
                        i = last;
                    }
                    else if (tokenPos === smlen) {
                        tokenPos = 0;
                        state = State.PreVariableSpace;
                        if (beginText >= 0 && beginText < last) {
                            current.pushData(source.slice(beginText, last));
                        }
                        beginText = -1;
                        fusionContext = tsfusionnode_1.TSFusionContextType.Local;
                    }
                    break;
                case State.PreVariableSpace:
                    if (isVariableSpace(c)) {
                        break;
                    }
                    const ct = TSFusionContextTypes[c];
                    if ((0, commons_1.$ok)(ct)) {
                        fusionContext = ct;
                        tokenPos = i + 1;
                        state = State.StartDecodingVariable;
                    }
                    else {
                        characterRole = variableCharacterRole(c);
                        if (characterRole === TSCharRole.UnsignificantStarterChar || characterRole === TSCharRole.StarterChar) {
                            tokenPos = i;
                            state = characterRole === TSCharRole.UnsignificantStarterChar ? State.StartDecodingVariable : State.DecodingVariable;
                        }
                        else {
                            throw new tserrors_1.TSError(`Malformed var : found forbidden initial variable character ${tohex(c)} ('${tochar(c)}') at position ${i}'.`, {
                                source: source,
                                character: c,
                                position: i,
                                state: state
                            });
                        }
                    }
                    break;
                case State.StartDecodingVariable:
                    characterRole = variableCharacterRole(c);
                    if (characterRole !== TSCharRole.UnsignificantStarterChar) {
                        if (characterRole === TSCharRole.StarterChar) {
                            state = State.DecodingVariable;
                        }
                        else {
                            throw new tserrors_1.TSError(`Malformed var : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i}.`, {
                                source: source,
                                character: c,
                                position: i,
                                state: state
                            });
                        }
                    }
                    break;
                case State.DecodingVariable:
                    const t = TSFusionNodeTypes[c];
                    if ((0, commons_1.$ok)(t) || isVariableSpace(c)) {
                        const variable = _data2ascii(source.slice(tokenPos, i));
                        current = current.pushVariable(variable, (0, commons_1.$ok)(t) ? t : tsfusionnode_1.TSFusionNodeType.Variable, fusionContext);
                        store_variable(this._variables, this._contextsVariables, variable, fusionContext);
                        state = State.PostVariableSpace;
                        fusionContext = tsfusionnode_1.TSFusionContextType.Local;
                        tokenPos = 0;
                    }
                    else if (c === em[0]) {
                        const variable = _data2ascii(source.slice(tokenPos, i));
                        current = current.pushVariable(variable, tsfusionnode_1.TSFusionNodeType.Variable, fusionContext);
                        store_variable(this._variables, this._contextsVariables, variable, fusionContext);
                        state = State.DecodingClosingVariable;
                        fusionContext = tsfusionnode_1.TSFusionContextType.Local;
                        i--;
                        tokenPos = 0;
                    }
                    else if (variableCharacterRole(c) === TSCharRole.ForbiddenChar) {
                        throw new tserrors_1.TSError(`Malformed var : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i}.`, {
                            source: source,
                            character: c,
                            position: i,
                            state: state
                        });
                    }
                    break;
                case State.PostVariableSpace:
                    if (c === em[0]) {
                        if (current.isContainerVariable) {
                            throw new tserrors_1.TSError(`Malformed template structure : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (Need to have a separator for a container variable).`, {
                                source: source,
                                character: c,
                                position: i,
                                neededChars: [sm[0]],
                                state: state
                            });
                        }
                        state = State.DecodingClosingVariable;
                        i--;
                        tokenPos = 0;
                        break;
                    }
                    else if (c === sepa[0]) {
                        if (!current.isContainerVariable) {
                            throw new tserrors_1.TSError(`Malformed template structure : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (Separator forbidden for non container variable).`, {
                                source: source,
                                character: c,
                                position: i,
                                neededChars: [em[0]],
                                state: state
                            });
                        }
                        state = State.DecodingSeparator;
                        i--;
                        tokenPos = 0;
                        break;
                    }
                    else if (!isVariableSpace(c)) {
                        throw new tserrors_1.TSError(`Malformed template structure : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (needed separator or ending mark start').`, {
                            source: source,
                            character: c,
                            position: i,
                            neededChars: [sepa[0], em[0]],
                            state: state
                        });
                    }
                    break;
                case State.DecodingSeparator:
                    comp = sepa[tokenPos++];
                    if (c !== comp) {
                        throw new tserrors_1.TSError(`Malformed template structure on variable separator. Unatended character  ${tohex(c)} ('${tochar(c)}') at position ${i} while decoding separator.`, {
                            source: source,
                            character: c,
                            position: i,
                            neededChars: [comp],
                            state: state
                        });
                    }
                    else if (tokenPos === sepalen) {
                        beginText = i + 1;
                        state = State.Text;
                    }
                    break;
                case State.DecodingClosingVariable:
                    comp = em[tokenPos++];
                    if (c !== comp) {
                        throw new tserrors_1.TSError(`Malformed template structure on closing variable. Unatended character  ${tohex(c)} ('${tochar(c)}') at position ${i} while decoding ending mark.`, {
                            source: source,
                            character: c,
                            position: i,
                            neededChars: [comp],
                            state: state
                        });
                    }
                    else if (tokenPos === emlen) {
                        current = current.parent;
                        state = State.Text;
                        beginText = i + 1;
                    }
                    break;
                case State.DecodeEndingMark:
                    comp = em[tokenPos++];
                    if (c !== comp) {
                        tokenPos = 0;
                        state = State.Text;
                        i = last;
                    }
                    else if (tokenPos === emlen) {
                        const parent = current.parent;
                        if ((0, commons_1.$ok)(parent)) {
                            if (beginText >= 0 && beginText < last) {
                                current.pushData(source.slice(beginText, last));
                            }
                            current = parent;
                            state = State.Text;
                            beginText = i + 1;
                        }
                        else {
                            throw new tserrors_1.TSError(`Malformed template structure. We have found a supernumerary ending mark at position ${i}.`, {
                                source: source,
                                position: i - emlen + 1,
                                endingMark: em,
                                state: state
                            });
                        }
                    }
                    break;
            }
            i++;
        }
        if (debug) {
            (0, utils_1.$logterm)(`&0&g--STOP--&w: &rEOF&x, state:&o${state}&x, current: &p${current.label}&0`);
        }
        if ((0, commons_1.$ok)(current.parent)) {
            throw new tserrors_1.TSError(`Malformed template structure. Template is not properly closed.`, {
                source: source,
                state: state
            });
        }
        else if (state === State.Text || state === State.OpenVariable || state === State.DecodeEndingMark) {
            if (beginText >= 0 && beginText < i) {
                current.pushData(source.slice(beginText, i));
            }
        }
        else {
            throw new tserrors_1.TSError(`Malformed template structure. Template is not properly closed (automat in wrong state '${state}').`, {
                source: source,
                state: state
            });
        }
        this.startingMark = sm;
        this.endingMark = em;
        this.separator = sepa;
        this.source = new tsdata_1.TSData(source);
        this._treeRoot = root;
        this._capacity = (0, commons_1.$capacityForCount)(this.source.length * 1.25);
        if ((0, commons_1.$ok)(opts.procedures)) {
            this._procedures = opts.procedures;
        }
    }
    get capacity() { return this._capacity; }
    set capacity(n) {
        const c = (0, commons_1.$capacityForCount)(n);
        if (c > this._capacity) {
            this._capacity = (this._capacity + (0, commons_1.$capacityForCount)(n));
        }
    }
    variablesForType(type) {
        const a = this._contextsVariables.get(type);
        return (0, commons_1.$ok)(a) && a.size > 0 ? Array.from(a) : [];
    }
    isVariableOfType(s, type) {
        if (!this.isVariable(s)) {
            return false;
        }
        const a = this._contextsVariables.get(type);
        return (0, commons_1.$ok)(a) && a.has(s);
    }
    get variables() { return Array.from(this._variables.keys()); }
    get userVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.User); }
    get globalVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.Global); }
    get systemVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.System); }
    get localVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.Local); }
    get rootVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.Root); }
    get procedures() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.Procedure); }
    isVariable(s) { return (0, commons_1.$ok)(this._variables.get(s)); }
    isUserVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.User); }
    isGlobalVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.Global); }
    isSystemVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.System); }
    isLocalVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.Local); }
    isRootVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.Root); }
    isProcedure(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.Procedure); }
}
exports.TSFusionTemplate = TSFusionTemplate;
TSFusionTemplate.DefaultStartingMark = '{{';
TSFusionTemplate.DefaultSeparator = ':';
TSFusionTemplate.DefaultEndingMark = '}}';
class TSStringTemplate extends TSFusionTemplate {
    static fromString(source, opts) {
        if (!(0, commons_1.$length)(source)) {
            return null;
        }
        try {
            return new TSStringTemplate(source, opts);
        }
        catch (e) {
            if (opts === null || opts === void 0 ? void 0 : opts.debugParsing) {
                (0, utils_1.$logterm)('&0\n&R&w Data Parsing did encounter error &0\n&c' + (0, utils_1.$inspect)(e) + '\n&o' + e.stack + "&0");
            }
            return null;
        }
    }
    constructor(source, options) {
        super(TSStringTemplate.__stringCharset.dataFromString(source), options);
        this._outputString = undefined;
    }
    pushFusionData(data) { this._outputString += TSStringTemplate.__stringCharset.stringFromData(data); }
    pushFusionValue(value) { this._outputString += (0, commons_1.$isstring)(value) ? value : ((0, commons_1.$isobject)(value) && 'toString' in value ? value.toString() : `${value}`); }
    fusionWithDataContext(data, globalContext, errors) {
        this._outputString = '';
        return this._treeRoot.fusion(this, data, globalContext, this._procedures, errors) ? this._outputString : null;
    }
}
exports.TSStringTemplate = TSStringTemplate;
TSStringTemplate.__stringCharset = tscharset_1.TSCharset.utf8Charset();
class TSGenericDataTemplate extends TSFusionTemplate {
    constructor() {
        super(...arguments);
        this._outputData = undefined;
    }
    pushFusionData(data) { var _a; (_a = this._outputData) === null || _a === void 0 ? void 0 : _a.appendBytes(data); }
    fusionWithDataContext(data, globalContext, errors) {
        this._outputData = new tsdata_1.TSData(this._capacity);
        return this._treeRoot.fusion(this, data, globalContext, this._procedures, errors) ? this._outputData : null;
    }
}
exports.TSGenericDataTemplate = TSGenericDataTemplate;
class TSHTMLTemplate extends TSGenericDataTemplate {
    static fromHTMLData(src, opts) {
        const source = src instanceof ArrayBuffer ? (0, data_1.$bufferFromArrayBuffer)(src) : src;
        if (!(0, commons_1.$length)(source)) {
            return null;
        }
        try {
            return new TSHTMLTemplate(source, opts);
        }
        catch (e) {
            if (opts === null || opts === void 0 ? void 0 : opts.debugParsing) {
                (0, utils_1.$logterm)('&0\n&R&w Data Parsing did encounter error &0\n&c' + (0, utils_1.$inspect)(e) + '\n&o' + e.stack + "&0");
            }
            return null;
        }
    }
    constructor(src, opts) {
        const source = src instanceof ArrayBuffer ? (0, data_1.$bufferFromArrayBuffer)(src) : src;
        super(_parseHTML(source, opts), opts);
    }
    pushFusionValue(value) {
        var _a;
        let s = undefined;
        if ((0, commons_1.$ok)(value)) {
            if ((0, commons_1.$isstring)(value)) {
                s = value.toHTML();
            }
            else if ((0, commons_1.$isnumber)(value)) {
                s = '' + value;
            }
            else if ((0, commons_1.$isobject)(value)) {
                if ('toHTML' in value) {
                    s = value.toHTML();
                }
                else if ('toString' in value) {
                    s = (value.toString()).toHTML();
                }
                else {
                    s = `${value}`.toHTML();
                }
            }
            else {
                s = `${value}`.toHTML();
            }
        }
        if ((0, commons_1.$length)(s)) {
            (_a = this._outputData) === null || _a === void 0 ? void 0 : _a.appendBytes(_ascii2data(s));
        }
    }
}
exports.TSHTMLTemplate = TSHTMLTemplate;
class TSDataTemplate extends TSGenericDataTemplate {
    static fromData(src, encoding, opts) {
        const source = src instanceof ArrayBuffer ? (0, data_1.$bufferFromArrayBuffer)(src) : src;
        if (!(0, commons_1.$length)(source)) {
            return null;
        }
        const charset = (0, tscharset_1.$charset)(encoding);
        if (TSFusionForbiddenEncodings.includes(charset.name)) {
            if (opts === null || opts === void 0 ? void 0 : opts.debugParsing) {
                (0, utils_1.$logterm)(`&0\n&R&w Data Parsing cannot use encoding: '${charset.name}'&0`);
            }
            return null;
        }
        try {
            return new TSDataTemplate(source, charset, opts);
        }
        catch (e) {
            if (opts === null || opts === void 0 ? void 0 : opts.debugParsing) {
                (0, utils_1.$logterm)('&0&R&w Data Parsing did encounter error &0\n&c' + (0, utils_1.$inspect)(e) + '\n&o' + e.stack + "&0");
            }
            return null;
        }
    }
    constructor(source, charset, opts) {
        super(source, opts);
        this._charset = charset;
    }
    get charset() { return this._charset; }
    pushFusionValue(value) {
        var _a;
        const s = (0, commons_1.$isstring)(value) ? value : ((0, commons_1.$isobject)(value) && 'toString' in value ? value.toString() : `${value}`);
        (_a = this._outputData) === null || _a === void 0 ? void 0 : _a.appendString(s, this._charset);
    }
}
exports.TSDataTemplate = TSDataTemplate;
const ASCIICharset = tscharset_1.TSCharset.asciiCharset();
function _ascii2data(s) { return ASCIICharset.bytesFromString(s); }
function _data2ascii(d, start, end) { return ASCIICharset.stringFromData(d, start, end); }
function _parseHTML(source, options) {
    const startingMark = (0, commons_1.$length)(options === null || options === void 0 ? void 0 : options.startingMark) ? options.startingMark : TSFusionTemplate.DefaultStartingMark;
    const endingMark = (0, commons_1.$length)(options === null || options === void 0 ? void 0 : options.endingMark) ? options.endingMark : TSFusionTemplate.DefaultEndingMark;
    const separator = (0, commons_1.$length)(options === null || options === void 0 ? void 0 : options.separator) ? options.separator : TSFusionTemplate.DefaultSeparator;
    const debug = options === null || options === void 0 ? void 0 : options.debugParsing;
    if (debug) {
        (0, utils_1.$logterm)(`&0\n&P&w _parseHTML() is in debug mode &0`);
        (0, utils_1.$logterm)(`&0&x - &c starting mark&x:        &p${startingMark}&0`);
        (0, utils_1.$logterm)(`&0&x - &c ending mark&x:          &p${endingMark}&0`);
        (0, utils_1.$logterm)(`&0&x - &c separator&x:            &p${separator}&0`);
    }
    const sm = _ascii2data(startingMark);
    const em = _ascii2data(endingMark);
    const sp = _ascii2data(separator);
    let State;
    (function (State) {
        State["Text"] = "text";
        State["Start"] = "start";
        State["CloserElement"] = "closer-element";
        State["Element"] = "element";
        State["Inside"] = "inside";
        State["Tag"] = "tag";
        State["AfterTag"] = "after-tag";
        State["AfterEqual"] = "after-equal";
        State["Quote"] = "quote";
        State["AfterQuote"] = "after-quote";
        State["Value"] = "value";
        State["Slash"] = "slash";
    })(State || (State = {}));
    ;
    let Element;
    (function (Element) {
        Element["Unknown"] = "";
        Element["Fusion"] = "FUSION";
        Element["EndFusion"] = "/FUSION";
    })(Element || (Element = {}));
    const assign = {
        'FUSION': ['TYPE', 'PATH', 'CONTEXT'],
        '/FUSION': []
    };
    const types = {
        'IF': [QUESTION_MARK],
        'TEST': [QUESTION_MARK],
        'QUESTION': [QUESTION_MARK],
        '?': [QUESTION_MARK],
        'IFNOT': [EXCLAMATION_MARK],
        'NOT': [EXCLAMATION_MARK],
        '!TEST': [QUESTION_MARK],
        '!': [EXCLAMATION_MARK],
        'ENUM': [SHARP],
        'ENUMERATION': [SHARP],
        'LIST': [SHARP],
        'ARRAY': [SHARP],
        '#': [SHARP],
        'VALUE': [],
        'PATH': [],
        'VAR': [],
        'VARIABLE': [],
        '': []
    };
    const contexts = {
        'GLOBAL': [DOLLAR],
        '$': [DOLLAR],
        'DICT': [DOLLAR],
        'DICTIONARY': [DOLLAR],
        'ROOT': [AT],
        'INITIAL': [AT],
        'TEMPLATE': [AT],
        '@': [AT],
        'SYS': [UNDERSCORE],
        'SYSTEM': [UNDERSCORE],
        '_': [UNDERSCORE],
        'PROC': [ASTERISK],
        'PROCEDURE': [ASTERISK],
        'FN': [ASTERISK],
        'FUNC': [ASTERISK],
        'FUNCTION': [ASTERISK],
        'CALL': [ASTERISK],
        '*': [ASTERISK],
        'USER': [PLUS],
        'USER-DEFINED': [PLUS],
        'PARAM': [PLUS],
        'PARAMETER': [PLUS],
        '+': [PLUS],
        'CURRENT': [DOT],
        'LOCAL': [DOT],
        '.': [DOT],
        '': [DOT]
    };
    const src = source instanceof tsdata_1.TSData ? source.mutableBuffer : source;
    const len = src.length;
    let target = new tsdata_1.TSData(len);
    let state = State.Text;
    let i = 0;
    let last = 0;
    let lastQuote = 0;
    let insideLast = 0;
    let tag = '';
    let element = Element.Unknown;
    let values = {};
    let isCloser = false;
    function upper(c) { return c & 0xdf; }
    function isNameStarterChar(c) { c = upper(c); return c >= A && c <= Z; }
    function isNameChar(c) { return c === DOT || c === COLON || c === UNDERSCORE || c === MINUS || (c >= ZERO && c <= NINE) || isNameStarterChar(c); }
    function isWhiteSpace(c) { return c === TAB || c === SPACE || c === LF || c === CR; }
    function isQuote(c) { return c === QUOTE || c === DOUBLE_QUOTE; }
    function name(start, end) { return _data2ascii(src, start, end).toUpperCase(); }
    function registerValue(start, end) {
        var _a;
        if ((_a = assign[element]) === null || _a === void 0 ? void 0 : _a.includes(tag)) {
            values[tag] = src.slice(start, end);
        }
        tag = '';
    }
    function reset() {
        state = State.Text;
        element = Element.Unknown;
        tag = '';
        last = 0;
        values = {};
        isCloser = false;
    }
    function back() {
        target.appendByte(LT);
        i = last;
        reset();
    }
    function endElement() {
        switch (element) {
            case Element.Fusion:
                let p = values['PATH'];
                if (!(0, commons_1.$length)(p)) {
                    back();
                    return;
                }
                let c = contexts[_data2ascii(values['CONTEXT']).toUpperCase()];
                let t = types[_data2ascii(values['TYPE']).toUpperCase()];
                target.appendBytes(sm);
                target.appendBytes(c);
                target.appendBytes(p);
                if (t.length) {
                    target.appendBytes(t);
                    target.appendBytes(sp);
                }
                else {
                    target.appendBytes(em);
                }
                reset();
                return;
            case Element.EndFusion:
                target.appendBytes(em);
                reset();
                return;
            default:
                target.appendBytes(src, last, i);
                target.appendByte(GT);
                break;
        }
        reset();
    }
    while (i < len) {
        const c = src[i];
        if (debug) {
            (0, utils_1.$logterm)(`&0&j${i.fpad(8, '_')}&w: &x0x&p${c.toString(16)}&x (&b'&c${String.fromCharCode(c)}&b'&x), state:&p${state}&x, last:&y${last}&0`);
        }
        switch (state) {
            case State.Text:
                if (c === LT) {
                    state = State.Start;
                    last = i;
                }
                else {
                    target.appendByte(c);
                }
                break;
            case State.Start:
                if (c === SLASH) {
                    state = State.CloserElement;
                    isCloser = true;
                }
                else if (isNameStarterChar(c)) {
                    state = State.Element;
                }
                else {
                    back();
                }
                break;
            case State.CloserElement:
                if (isNameStarterChar(c)) {
                    state = State.Element;
                }
                else {
                    back();
                }
                break;
            case State.Element:
                if (c === GT || isWhiteSpace(c)) {
                    const elementName = name(last + 1, i).toUpperCase();
                    if (!(0, commons_1.$ok)(assign[elementName])) {
                        back();
                    }
                    else {
                        element = elementName;
                        if (c === GT) {
                            endElement();
                        }
                        else {
                            state = State.Inside;
                        }
                    }
                }
                else if (!isNameChar(c)) {
                    back();
                }
                break;
            case State.Inside:
                if (c === GT) {
                    endElement();
                }
                else if (isCloser && !isWhiteSpace(c)) {
                    back();
                }
                else if (c === SLASH) {
                    state = State.Slash;
                }
                else if (isNameStarterChar(c)) {
                    insideLast = i;
                    state = State.Tag;
                }
                else if (!isWhiteSpace(c)) {
                    back();
                }
                break;
            case State.Tag:
                if (c === EQ || isWhiteSpace(c)) {
                    tag = name(insideLast, i);
                    state = c === EQ ? State.AfterEqual : State.AfterTag;
                }
                else if (!isNameChar(c)) {
                    back();
                }
                break;
            case State.AfterTag:
                if (c === EQ) {
                    state = State.AfterEqual;
                }
                else if (!isWhiteSpace(c)) {
                    back();
                }
                break;
            case State.AfterEqual:
                if (isQuote(c)) {
                    insideLast = i + 1;
                    lastQuote = c;
                    state = State.Quote;
                }
                else if (isNameStarterChar(c)) {
                    insideLast = i;
                    state = State.Value;
                }
                else if (!isWhiteSpace(c)) {
                    back();
                }
                break;
            case State.Value:
                if (c === GT || c === SLASH || isWhiteSpace(c)) {
                    registerValue(insideLast, i);
                    if (c === GT) {
                        endElement();
                    }
                    else if (c === SLASH) {
                        state = State.Slash;
                    }
                    else {
                        state = State.Inside;
                    }
                }
                else if (!isNameChar(c)) {
                    back();
                }
                break;
            case State.Quote:
                if (c === lastQuote) {
                    registerValue(insideLast, i);
                    state = State.AfterQuote;
                }
                break;
            case State.AfterQuote:
                if (isWhiteSpace(c)) {
                    state = State.Inside;
                }
                else if (c === SLASH) {
                    state = State.Slash;
                }
                else if (c === GT) {
                    endElement();
                }
                else {
                    back();
                }
                break;
            case State.Slash:
                if (c === GT) {
                    endElement();
                }
                else if (!isWhiteSpace(c)) {
                    back();
                }
                break;
        }
        i++;
    }
    if (state !== State.Text) {
        target.appendByte(LT);
        target.appendBytes(src, last, i);
        target.appendByte(GT);
    }
    if (debug) {
        (0, utils_1.$logterm)(`&0&j--STOP--&w: &rEOF&x, state:&p${state}&x, last:&p${last}&0`);
    }
    return target;
}
//# sourceMappingURL=tsfusion.js.map