"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSFusionTreeNode = exports.TSFusionNodeType = exports.TSFusionContextType = void 0;
const commons_1 = require("./commons");
const strings_1 = require("./strings");
const tsdate_1 = require("./tsdate");
const tsdefaults_1 = require("./tsdefaults");
const tserrors_1 = require("./tserrors");
const tsobject_1 = require("./tsobject");
const utils_1 = require("./utils");
var TSFusionContextType;
(function (TSFusionContextType) {
    TSFusionContextType["Local"] = "local";
    TSFusionContextType["Global"] = "global";
    TSFusionContextType["Root"] = "root";
    TSFusionContextType["System"] = "system";
    TSFusionContextType["Procedure"] = "procedure";
    TSFusionContextType["User"] = "user";
})(TSFusionContextType = exports.TSFusionContextType || (exports.TSFusionContextType = {}));
;
var TSFusionNodeType;
(function (TSFusionNodeType) {
    TSFusionNodeType["Root"] = "root";
    TSFusionNodeType["Data"] = "data";
    TSFusionNodeType["Variable"] = "variable";
    TSFusionNodeType["Test"] = "test-variable";
    TSFusionNodeType["NegativeTest"] = "nagative-test-variable";
    TSFusionNodeType["Enumeration"] = "enumeration-variable";
})(TSFusionNodeType = exports.TSFusionNodeType || (exports.TSFusionNodeType = {}));
class TSFusionTreeNode {
    static root() { return new TSFusionTreeNode(TSFusionNodeType.Root); }
    constructor(type, value, parent) {
        this.contextType = TSFusionContextType.Local;
        this._localContext = {};
        this.type = type;
        if ((0, commons_1.$ok)(value)) {
            this.value = value;
        }
        if ((0, commons_1.$ok)(parent)) {
            this.parent = parent;
        }
    }
    get label() {
        switch (this.type) {
            case TSFusionNodeType.Root: return 'Root';
            case TSFusionNodeType.Data: return 'Data';
            default: return this.value;
        }
    }
    get isContainerVariable() {
        return this.type === TSFusionNodeType.Test || this.type === TSFusionNodeType.NegativeTest || this.type === TSFusionNodeType.Enumeration;
    }
    pushData(data) {
        if (this.type === TSFusionNodeType.Data) {
            throw new tserrors_1.TSError('Impossible to add subnode to a text FusionTreeNode', { text: this.value });
        }
        if (data.length) {
            if (!(0, commons_1.$ok)(this.nodes)) {
                this.nodes = [];
            }
            this.nodes.push(new TSFusionTreeNode(TSFusionNodeType.Data, data, this));
        }
    }
    pushVariable(name, type, contextType) {
        if (this.type === TSFusionNodeType.Data) {
            throw new tserrors_1.TSError('Impossible to add subnode to a text FusionTreeNode', { text: this.value });
        }
        if (!(0, commons_1.$length)(name)) {
            throw new tserrors_1.TSError(`Impossible to add an unamed fusion variable to ${this.label} variable.`);
        }
        if (type === TSFusionNodeType.Root || type === TSFusionNodeType.Data) {
            throw new tserrors_1.TSError(`Impossible to add fusion variable '${name}' as a ${type} variable`, { name: name, type: type });
        }
        if (!(0, commons_1.$ok)(this.nodes)) {
            this.nodes = [];
        }
        let node = new TSFusionTreeNode(type, name, this);
        if ((0, commons_1.$ok)(contextType)) {
            node.contextType = contextType;
        }
        node._pathCache = name.split('.');
        this.nodes.push(node);
        return node;
    }
    fusion(template, data, globalContext, procedures = {}, errors) {
        if (this.type !== TSFusionNodeType.Root) {
            throw new tserrors_1.TSError('Impossible invoque fusion() method on non-root FusionNodeType', { label: this.label, type: this.type });
        }
        try {
            this._cleanLocalContexts();
            const defaults = tsdefaults_1.TSDefaults.defaults();
            this._fusion(template, globalContext, [data], [this._localContext], {
                zulu: tsdate_1.TSDate.zulu(),
                date: new tsdate_1.TSDate(),
                language: defaults.language(),
                languages: defaults.managedLanguages(),
                currency: defaults.currency(),
                locales: defaults.locales(),
                index: 0,
                position: 1,
                remaining: 0,
                count: 1
            }, procedures, errors);
            return true;
        }
        catch (e) {
            if ((0, commons_1.$ok)(errors)) {
                errors.push('Fusion did encounter error:');
                (0, strings_1.$lines)((0, utils_1.$inspect)(e)).forEach(l => errors === null || errors === void 0 ? void 0 : errors.push(l));
                errors.push('STACK:');
                (0, strings_1.$lines)(e.stack).forEach(l => errors === null || errors === void 0 ? void 0 : errors.push(l));
            }
        }
        return false;
    }
    _cleanLocalContexts() {
        var _a;
        this._localContext = {};
        (_a = this.nodes) === null || _a === void 0 ? void 0 : _a.forEach(n => n._cleanLocalContexts());
    }
    _fusion(template, globalContext, stack, localStack, systemContext, procedures, errors) {
        var _a;
        const debug = (0, commons_1.$ok)(errors);
        const count = (0, commons_1.$count)(this.nodes);
        for (let i = 0; i < count; i++) {
            const node = this.nodes[i];
            if (node.type === TSFusionNodeType.Data) {
                template.pushFusionData(node.value);
            }
            else {
                let v = undefined;
                if (node.contextType === TSFusionContextType.Procedure) {
                    if (((_a = node._pathCache) === null || _a === void 0 ? void 0 : _a.length) === 1 && (0, commons_1.$length)(node.value) > 0) {
                        const proc = procedures[node.value];
                        if ((0, commons_1.$isfunction)(proc)) {
                            try {
                                v = proc(stack.last(), stack.first(), this._localContext, globalContext, Object.assign({}, systemContext));
                            }
                            catch (e) {
                                if (debug) {
                                    errors.push(`!ERROR!: Procedure ${node.value}() execution did fail`);
                                }
                                v = undefined;
                            }
                        }
                        else if (debug) {
                            errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: Procedure ${node.value}() does not exist`);
                        }
                    }
                    else if (debug) {
                        errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: Malformed procedure name ${node.value}() (undefined, empty or with '.' character in it)`);
                    }
                }
                else {
                    let target = [];
                    let dataType = 'unknown target';
                    switch (node.contextType) {
                        case TSFusionContextType.Root:
                            target = [stack.first()];
                            dataType = 'root';
                            break;
                        case TSFusionContextType.Local:
                            target = localStack;
                            dataType = 'local context';
                            break;
                        case TSFusionContextType.Global:
                            target = [globalContext];
                            dataType = 'global context';
                            break;
                        case TSFusionContextType.User:
                            target = [globalContext];
                            dataType = 'user global context';
                            break;
                        case TSFusionContextType.System:
                            target = [systemContext];
                            dataType = 'system context';
                            break;
                    }
                    v = _valueForKeyPath(target, node._pathCache, dataType, false, errors);
                    if (node.contextType === TSFusionContextType.Local && !(0, commons_1.$ok)(v)) {
                        v = _valueForKeyPath(stack, node._pathCache, 'target object', true, errors);
                    }
                }
                switch (node.type) {
                    case TSFusionNodeType.Variable:
                        if ((0, commons_1.$ok)(v)) {
                            template.pushFusionValue(v);
                        }
                        else {
                            errors === null || errors === void 0 ? void 0 : errors.push(`WARNING: value returned for '${node.value}' was undefined or null and should have result in data to merge in fusion.`);
                        }
                        break;
                    case TSFusionNodeType.Test:
                        if (v) {
                            node._fusion(template, globalContext, stack, localStack, systemContext, procedures, errors);
                        }
                        break;
                    case TSFusionNodeType.NegativeTest:
                        if (!v) {
                            node._fusion(template, globalContext, stack, localStack, systemContext, procedures, errors);
                        }
                        break;
                    case TSFusionNodeType.Enumeration:
                        if ((0, commons_1.$ok)(v)) {
                            const a = (0, tsobject_1.$iscollection)(v) ? v.getItems() : ('fusionEnumeration' in v ? v.fusionEnumeration() : [v]);
                            const acount = (0, commons_1.$length)(a);
                            for (let j = 0; j < acount; j++) {
                                node._fusion(template, globalContext, [...stack, a[j]], [...localStack, a[j]._localContext], Object.assign(Object.assign({}, systemContext), { count: acount, index: j, position: j + 1, remaining: count - j - 1 }), procedures, errors);
                            }
                        }
                        break;
                    default: break;
                }
            }
        }
    }
}
exports.TSFusionTreeNode = TSFusionTreeNode;
function _valueForKeyPath(stack, keyPath = [], dataType, unknownAsError, errors) {
    const scount = stack.length;
    const kcount = keyPath.length;
    if (!kcount) {
        errors === null || errors === void 0 ? void 0 : errors.push('ERROR: empty key path');
        return undefined;
    }
    let i = 0;
    for (; i < kcount && keyPath[i].length === 0; i++) { }
    if (i >= scount) {
        errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: key path '${keyPath.join('.')}' did underflow the stack.`);
        return undefined;
    }
    let v = stack[scount - i - 1];
    for (; i < kcount && (0, commons_1.$ok)(v); i++) {
        const m = keyPath[i];
        if (!m.length) {
            errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: key path '${keyPath.join('.')}' contains internal '..'.`);
            return undefined;
        }
        if (m === 'self') {
            continue;
        }
        if ((0, commons_1.$isobject)(v) && !(m in v)) {
            errors === null || errors === void 0 ? void 0 : errors.push(`${unknownAsError ? '!ERROR!' : 'WARNING'}: unknown method or preperty '${m}' in ${dataType}.`);
            return undefined;
        }
        let res = v[m];
        if ((0, commons_1.$isfunction)(res)) {
            if (res.length > 0) {
                errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: method ${m}() is not an sinple accessor in ${dataType}.`);
                return undefined;
            }
            try {
                v = res.call(v);
            }
            catch (e) {
                errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: method ${m}() execution did fail on ${dataType}`);
                return undefined;
            }
        }
        else {
            v = res;
        }
        if (typeof v === 'number' && isNaN(v)) {
            errors === null || errors === void 0 ? void 0 : errors.push(`WARNING: key path '${keyPath.join('.')}' would return NaN which was transformed to null.`);
            return null;
        }
    }
    return v;
}
Set.prototype.fusionEnumeration = function fusionEnumeration() { return Array.from(this); };
Map.prototype.fusionEnumeration = function fusionEnumeration() {
    const ret = [];
    this.forEach((value, key) => ret.push({ key: key, value: value }));
    return ret;
};
//# sourceMappingURL=tsfusionnode.js.map